/*
 * Copyright (C) 2000 Robert Fitzsimons
 */

#include "datatypes.h"
#include "unicode.h"
#include "classlib.h"
#include "classfile.h"
#include "classdata.h"
#include "hashmap.h"
#include "debug.h"

#include <stdlib.h>

ClassData* classdata_build(kuint class_handle, ClassFile* classfile);
void classdata_build_fielddata(ClassFile* classfile, ClassData* classdata);
void classdata_build_methoddata(ClassFile* classfile, ClassData* classdata);
void classdata_build_interfacedata(ClassFile* classfile, ClassData* classdata);
ConstantInfo* constant_resolve(ClassFile* classfile, kuint index);
void parse_field_descriptor(kuint descriptor_handle, kuint* flags);
void parse_method_descriptor(kuint descriptor_handle, kuint* flags);

extern struct ClassLib classlib_array[];
HashMap* classdata_hashmap;

void init_classdata() {
	classdata_hashmap = hashmap_create();
}

ClassData* find_classdata(kuint class_handle) {
	Unicode* classname = NULL;
	ClassData* classdata = NULL;
	ClassFile* classfile = NULL;

	classdata = (ClassData*)hashmap_get(classdata_hashmap, class_handle);

	if(classdata) {
		return classdata;
	}

	classname = unicode_get(class_handle);

	if(classname->array[0] != '[') {
		struct ClassLib* classlib = NULL;
		kuint index = 0;

		while((classlib = &classlib_array[index++])->hash != 0) {
			if(classlib->hash == class_handle) {
				classfile = classfile_build(classlib->buffer, classlib->buffer_length);
				// classfile_print(classfile);
				break;
			}
		}
	} else {
		FIXME();
	}

	if(!classfile) {
		FIXME();
	}

	classdata = classdata_build(class_handle, classfile);

	classfile_free(classfile);

	if(!classdata) {
		FIXME();
	}

	print_classdata(classdata);

	return classdata;
}

ClassData* classdata_build(kuint class_handle, ClassFile* classfile) {
	ClassData* classdata = NULL;

	classdata = calloc(1, sizeof(ClassData));
	classdata->class_handle = class_handle;

	hashmap_put(classdata_hashmap, class_handle, classdata);

	if(classfile->super_class) {
		ConstantInfo* ci = constant_resolve(classfile, classfile->super_class);
		classdata->super_classdata = find_classdata(ci->class.unicode->hash);
	}

	classdata->flags = classfile->access_flags << 16;

	classdata_build_fielddata(classfile, classdata);
	classdata_build_methoddata(classfile, classdata);
	classdata_build_interfacedata(classfile, classdata);

	return classdata;
}

void classdata_build_fielddata(ClassFile* classfile, ClassData* classdata) {
	kuint index = 0;
	kuint count = 0;

	if(classdata->super_classdata != NULL) {
		if(classfile->field_count) {
			classdata->fielddata_count = classdata->super_classdata->fielddata_count + classfile->field_count;
			classdata->fielddata = calloc(classdata->fielddata_count, sizeof(FieldData*));

			memcpy(classdata->fielddata, classdata->super_classdata->fielddata, classdata->super_classdata->fielddata_count * sizeof(FieldData*));
			count = classdata->super_classdata->fielddata_count;

			classdata->class_reference_data_length = classdata->super_classdata->class_reference_data_length;
			classdata->class_primitive_data_length = classdata->super_classdata->class_primitive_data_length;
			classdata->object_reference_data_length = classdata->super_classdata->object_reference_data_length;
			classdata->object_primitive_data_length = classdata->super_classdata->object_primitive_data_length;
		} else {
			classdata->fielddata_count = classdata->super_classdata->fielddata_count;
			classdata->fielddata = classdata->super_classdata->fielddata;

			classdata->class_reference_data_length = classdata->super_classdata->class_reference_data_length;
			classdata->class_primitive_data_length = classdata->super_classdata->class_primitive_data_length;
			classdata->object_reference_data_length = classdata->super_classdata->object_reference_data_length;
			classdata->object_primitive_data_length = classdata->super_classdata->object_primitive_data_length;

			return;
		}
	} else {
		if(classfile->field_count) {
			classdata->fielddata_count = classfile->field_count;
			classdata->fielddata = calloc(classdata->fielddata_count, sizeof(FieldData*));
		} else {
			return;
		}
	}

	for(index = 0; index < classfile->field_count; index++) {
		FieldInfo* fi = &classfile->field[index];
		FieldData* fd = calloc(1, sizeof(FieldData));

		fd->classdata = classdata;
		fd->flags = fi->access_flags << 16;
		fd->name_handle = (constant_resolve(classfile, fi->name_index))->unicode.unicode->hash;
		fd->descriptor_handle = (constant_resolve(classfile, fi->descriptor_index))->unicode.unicode->hash;

		parse_field_descriptor(fd->descriptor_handle, &fd->flags);

		if(!(fd->flags & FLAG_STATIC)) {
			if(!(fd->flags & FLAG_TYPE_REFERENCE)) {
				fd->offset = classdata->object_primitive_data_length;
				classdata->object_primitive_data_length += ((fd->flags & FLAG_SIZE_MASK) == 8) ? 8 : 4;
			} else {
				fd->offset = classdata->object_reference_data_length;
				classdata->object_reference_data_length += (fd->flags & FLAG_SIZE_MASK);
			}
		} else {
			if(!(fd->flags & FLAG_TYPE_REFERENCE)) {
				fd->offset = classdata->class_primitive_data_length;
				classdata->class_primitive_data_length += ((fd->flags & FLAG_SIZE_MASK) == 8) ? 8 : 4;
			} else {
				fd->offset = classdata->class_reference_data_length;
				classdata->class_reference_data_length += (fd->flags & FLAG_SIZE_MASK);
			}
		}

		classdata->fielddata[count++] = fd;
	}
}

void classdata_build_methoddata(ClassFile* classfile, ClassData* classdata) {
	if(classdata->super_classdata != NULL) {
		kuint index = 0;
		kuint index2 = 0;
		kuint count = 0;
		kuint found = 0;

		for(index = 0; index < classfile->method_count; index++) {
			MethodInfo* mi = &classfile->method[index];

			mi->name_index = (constant_resolve(classfile, mi->name_index))->unicode.unicode->hash;
			mi->descriptor_index = (constant_resolve(classfile, mi->descriptor_index))->unicode.unicode->hash;
			mi->access_flags <<= 16;

			for(index2 = 0; index2 < classdata->super_classdata->methoddata_count; index2++) {
				MethodData* smd = classdata->super_classdata->methoddata[index2];

				if((mi->name_index == smd->name_handle) && (mi->descriptor_index == smd->descriptor_handle)) {
					found++;
					break;
				}
			}

			if(index2 >= classdata->super_classdata->methoddata_count) {
				count++;
			}
		}

		if(count || found) {
			classdata->methoddata_count = classdata->super_classdata->methoddata_count + count;
			classdata->methoddata = calloc(classdata->methoddata_count, sizeof(MethodData*));

			memcpy(classdata->methoddata, classdata->super_classdata->methoddata, classdata->super_classdata->methoddata_count * sizeof(MethodData*));
			count = classdata->super_classdata->methoddata_count;
		} else {
			classdata->methoddata_count = classdata->super_classdata->methoddata_count;
			classdata->methoddata = classdata->super_classdata->methoddata;

			return;
		}

		for(index = 0; index < classfile->method_count; index++) {
			MethodInfo* mi = &classfile->method[index];
			MethodData* md = calloc(1, sizeof(MethodData));

			md->classdata = classdata;
			md->name_handle = mi->name_index;
			md->descriptor_handle = mi->descriptor_index;
			md->flags = mi->access_flags;

			parse_method_descriptor(md->descriptor_handle, &md->flags);

			for(index2 = 0; index2 < classdata->super_classdata->methoddata_count; index2++) {
				MethodData* smd = classdata->super_classdata->methoddata[index2];

				if((mi->name_index == smd->name_handle) && (mi->descriptor_index == smd->descriptor_handle)) {
					classdata->methoddata[index2] = md;
					break;
				}
			}

			if(index2 >= classdata->super_classdata->methoddata_count) {
				classdata->methoddata[count] = md;
				count++;
			}
		}
	} else {
		kuint index = 0;

		if(classfile->method_count) {
			classdata->methoddata_count = classfile->method_count;
			classdata->methoddata = calloc(classdata->methoddata_count, sizeof(MethodData*));
		} else {
			return;
		}

		for(index = 0; index < classfile->method_count; index++) {
			MethodInfo* mi = &classfile->method[index];
			MethodData* md = calloc(1, sizeof(MethodData));

			md->classdata = classdata;
			md->name_handle = (constant_resolve(classfile, mi->name_index))->unicode.unicode->hash;
			md->descriptor_handle = (constant_resolve(classfile, mi->descriptor_index))->unicode.unicode->hash;
			md->flags = mi->access_flags << 16;

			parse_method_descriptor(md->descriptor_handle, &md->flags);

			classdata->methoddata[index] = md;
		}
	}
}

void classdata_build_interfacedata(ClassFile* classfile, ClassData* classdata) {
	if(classdata->super_classdata != NULL) {
		kuint index = 0;
		kuint index2 = 0;
		kuint count = 0;
		kuint found = 0;

		for(index = 0; index < classfile->interface_count; index++) {
			kuint ii = classfile->interface[index];

			classfile->interface[index] = ii = (constant_resolve(classfile, ii))->unicode.unicode->hash;

			for(index2 = 0; index2 < classdata->super_classdata->interfacedata_count; index2++) {
				kuint sii = classdata->super_classdata->interfacedata[index2]->classdata->class_handle;

				if(ii == sii) {
					found++;
					break;
				}
			}

			if(index2 >= classdata->super_classdata->interfacedata_count) {
				count++;
			}
		}

		if(count || found) {
			classdata->interfacedata_count = classdata->super_classdata->interfacedata_count + count;
			classdata->interfacedata = calloc(classdata->interfacedata_count, sizeof(InterfaceData*));

			memcpy(classdata->interfacedata, classdata->super_classdata->interfacedata, classdata->super_classdata->interfacedata_count * sizeof(InterfaceData*));
			count = classdata->super_classdata->interfacedata_count;
		} else {
			classdata->interfacedata_count = classdata->super_classdata->interfacedata_count;
			classdata->interfacedata = classdata->super_classdata->interfacedata;

			return;
		}

		for(index = 0; index < classfile->interface_count; index++) {
			kuint ii = classfile->interface[index];
			InterfaceData* id = calloc(1, sizeof(InterfaceData));

			id->classdata = find_classdata(ii);

			for(index2 = 0; index2 < classdata->super_classdata->interfacedata_count; index2++) {
				kuint sii = classdata->super_classdata->interfacedata[index2]->classdata->class_handle;

				if(ii == sii) {
					classdata->interfacedata[index2] = id;
					break;
				}
			}

			if(index2 >= classdata->super_classdata->interfacedata_count) {
				classdata->interfacedata[count] = id;
				count++;
			}
		}
	}
}

ConstantInfo* constant_resolve(ClassFile* classfile, kuint index) {
	ConstantInfo* ci = & classfile->constant[index];

	if(!(ci->info.tag & CONSTANT_Resolved)) {
		switch(ci->info.tag & CONSTANT_Info_Mask) {
			case CONSTANT_Info_Utf8:
				ci->unicode.tag = CONSTANT_Resolved | CONSTANT_Unicode;
				ci->unicode.unicode = unicode_put_utf8(ci->utf8_info.bytes, ci->utf8_info.length);
				break;
			case CONSTANT_Info_Class:
				ci->class.tag = CONSTANT_Resolved | CONSTANT_Class;
				ci->class.unicode = unicode_put_special(constant_resolve(classfile, ci->class_info.name_index)->unicode.unicode, 'L', ';');
				break;
			default:
				FIXME();
		}
	}

	return ci;
}

void parse_method_descriptor(kuint descriptor_handle, kuint* flags) {
	Unicode* descriptor = unicode_get(descriptor_handle);
	kuint c = 0;
	kuint index = 0;
	kuint16* array = descriptor->array;
	kuint count = 0;
	kuint return_type = 0;

	for(index = 1; (c = array[index]) != ')'; index++) {
		switch(c) {
			case '[':
				while((c = array[++index]) == '[') {
				}
				if(c != 'L') {
					count++;
					break;
				}
			case 'L':
				while((c = array[++index]) != ';') {
				}
				count++;
				break;
			case 'D':
			case 'J':
				count++;
			default:
				count++;
				break;
		}
	}

	switch(c = array[++index]) {
		case '[':
		case 'L':
			return_type = FLAG_RETURN_REFERENCE;
			break;
		case 'D':
		case 'J':
			return_type = FLAG_RETURN_VALUE64;
			break;
		case 'V':
			return_type = FLAG_RETURN_VOID;
			break;
		default:
			return_type = FLAG_RETURN_VALUE32;
			break;
	}

	if(!((*flags) & FLAG_STATIC)) {
		count++;
	}

	(*flags) |= count | return_type;
}

void parse_field_descriptor(kuint descriptor_handle, kuint* flags) {
	Unicode* descriptor = unicode_get(descriptor_handle);
	kuint size = 0;
	kuint type = 0;

	switch(descriptor->array[0]) {
		case '[':
		case 'L':
			size = sizeof(Object*);
			type = FLAG_TYPE_REFERENCE;
			break;
		case 'Z':
		case 'B':
			size = sizeof(kuint8);
			break;
		case 'C':
		case 'S':
			size = sizeof(kuint16);
			break;
		case 'F':
		case 'I':
			size = sizeof(kuint32);
			break;
		case 'D':
		case 'J':
			size = sizeof(kuint64);
			break;
	}

	(*flags) |= size | type;
}

void print_classdata(ClassData* classdata) {
	kuint index = 0;

	debug(
		"ClassData\t0x%08X\n",
		(kuint)classdata
	);

	debug("%08X\t", classdata->class_handle);
	debug_hash(classdata->class_handle, 1);

	if(classdata->super_classdata) {
		debug("%08X\t", classdata->super_classdata->class_handle);
		debug_hash(classdata->super_classdata->class_handle, 1);
	} else {
		debug("%08X\n", 0);
	}

	debug("%08X\n", classdata->flags);

	debug("%08X\n", classdata->methoddata_count);
	for(index = 0; index < classdata->methoddata_count; index++) {
		MethodData* md = classdata->methoddata[index];
		debug(
			"  %08X  0x%08X  %08X  %08X  %08X\n",
			index,
			md->classdata,
			md->name_handle,
			md->descriptor_handle,
			md->flags
		);
	}

	debug("%08X\n", classdata->fielddata_count);
	for(index = 0; index < classdata->fielddata_count; index++) {
		FieldData* fd = classdata->fielddata[index];
		debug(
			"  %08X  0x%08X  %08X  %08X  %08X  %08X\n",
			index,
			fd->classdata,
			fd->name_handle,
			fd->descriptor_handle,
			fd->flags,
			fd->offset
		);
	}

	debug("%08X\n", classdata->interfacedata_count);
	for(index = 0; index < classdata->interfacedata_count; index++) {
		InterfaceData* id = classdata->interfacedata[index];
		debug(
			"  %08X  0x%08X  %08X\n",
			index,
			id->classdata,
			id->classdata->class_handle
		);
	}

	debug(
		"%08X\n%08X\n%08X\n%08X\n",
		classdata->class_reference_data_length,
		classdata->class_primitive_data_length,
		classdata->object_reference_data_length,
		classdata->object_primitive_data_length
	);
}

