/*
 * Copyright (C) 2000 Robert Fitzsimons
 */

#include "basetypes.h"
#include "datatypes.h"
#include "hashmap.h"
#include "unicode.h"
#include "debug.h"

#include <stdlib.h>
#include <string.h>

void build_crc32_table();

#define read_utf8(u,ui) ({ \
	kuint __c = (u)[(ui)++]; \
	if(__c < 0xC0) { \
	} else if(__c < 0xE0) { \
		__c = (((__c & 0x1F) << 6) + ((u)[(ui)++] & 0x3F)); \
	} else { \
		__c = ((((__c & 0x0F) << 12) + (((u)[(ui)++] & 0x3F) << 6)) + ((u)[(ui)++] & 0x3F)); \
	} \
	__c; \
})
#define add_crc(crc, c) ({ \
	(crc) = ((crc) << 8) ^ crc32_table[((crc) >> 24) ^ ((c) & 0xFF)]; \
	(crc) = ((crc) << 8) ^ crc32_table[((crc) >> 24) ^ ((c) >> 8)]; \
})

HashMap* unicode_hashmap;
kuint32* crc32_table;

void init_unicode() {
	unicode_hashmap = hashmap_create();
	build_crc32_table();
}

Unicode* unicode_put_array(kuint16* array, kuint array_length) {
	kuint hash = unicode_hash_array(array, array_length);
	Unicode* unicode = hashmap_get(unicode_hashmap, hash);

	if(!unicode) {
		unicode = calloc(1, sizeof(Unicode) + (array_length * sizeof(kuint16)));

		unicode->hash = hash;
		unicode->length = array_length;
		memcpy(unicode->array, array, (array_length * sizeof(kuint16)));

		hashmap_put(unicode_hashmap, hash, unicode);
	}

	return unicode;
}

Unicode* unicode_put_utf8(kuint8* utf8, kuint utf8_length) {
	kuint new_length = 0;
	kuint hash = unicode_hash_utf8(utf8, utf8_length, &new_length);
	Unicode* unicode = hashmap_get(unicode_hashmap, hash);

	if(!unicode) {
		kuint index = 0;
		kuint utf8_index = 0;

		unicode = calloc(1, sizeof(Unicode) + (new_length * sizeof(kuint16)));

		unicode->hash = hash;
		unicode->length = new_length;

		while(utf8_index < utf8_length) {
			unicode->array[index++] = read_utf8(utf8, utf8_index);
		}

		hashmap_put(unicode_hashmap, hash, unicode);
	}

	return unicode;
}

Unicode* unicode_put_ascii(kuint8* ascii) {
	return unicode_put_utf8(ascii, strlen(ascii));
}

Unicode* unicode_put_special(Unicode* old_unicode, kuint start_char, kuint end_char) {
	kuint hash = unicode_hash_special(old_unicode, start_char, end_char);
	Unicode* unicode = hashmap_get(unicode_hashmap, hash);

	if(!unicode) {
		kuint length = old_unicode->length + 2;

		unicode = calloc(1, sizeof(Unicode) + (length * sizeof(kuint16)));

		unicode->hash = hash;
		unicode->length = length;
		memcpy(&unicode->array[1], old_unicode->array, (old_unicode->length * sizeof(kuint16)));

		unicode->array[0] = start_char;
		unicode->array[length - 1] = end_char;

		hashmap_put(unicode_hashmap, hash, unicode);
	}

	return unicode;
}

Unicode* unicode_get(kuint hash) {
	return hashmap_get(unicode_hashmap, hash);
}

kuint unicode_hash_array(kuint16* array, kuint array_length) {
	kuint crc = 0xFFFFFFFF;
	kuint buffer_length = array_length;
	kuint16* buffer = array;

	while(buffer_length-- > 0) {
		kuint c = (*(buffer++));
		add_crc(crc, (c & 0xFF));
	}

	return ~crc;
}

kuint unicode_hash_utf8(kuint8* utf8, kuint utf8_length, kuint* new_length) {
	kuint crc = 0xFFFFFFFF;
	kuint c = 0;
	kuint index = 0;
	kuint length = 0;

	while(index < utf8_length) {
		c = read_utf8(utf8, index);
		add_crc(crc, c);
		length++;
	}

	if(new_length) {
		(*new_length) = length;
	}

	return ~crc;
}

kuint unicode_hash_special(Unicode* unicode, kuint start_char, kuint end_char) {
	kuint crc = 0xFFFFFFFF;
	kuint buffer_length = unicode->length;
	kuint16* buffer = unicode->array;

	add_crc(crc, start_char);

	while(buffer_length--) {
		kuint c = (*(buffer++));
		add_crc(crc, (c & 0xFF));
	}

	add_crc(crc, end_char);

	return ~crc;
}

kuint8* unicode_to_utf8(Unicode* unicode, kuint* utf8_length) {
	kuint index = 0;
	kuint length = 0;
	kuint c = 0;
	kuint8* utf8 = NULL;

	for(index = 0; index < unicode->length; index++) {
		c = unicode->array[index];
		if(c < 0x80) {
			length++;
		} else if(c < 0x800) {
			length += 2;
		} else {
			length += 3;
		}
	}

	utf8 = calloc(length + 1, sizeof(kuint8));
	length = 0;

	for(index = 0; index < unicode->length; index++) {
		c = unicode->array[index];
		if(c < 0x80) {
			utf8[length++] = c;
		} else if(c < 0x800) {
			utf8[length++] = (0xC0 | ((c >> 6) & 0x1F));
			utf8[length++] = (0x80 | ((c >> 0) & 0x3F));
		} else {
			utf8[length++] = (0xE0 | ((c >> 12) & 0x0F));
			utf8[length++] = (0x80 | ((c >> 6) & 0x3F));
			utf8[length++] = (0x80 | ((c >> 0) & 0x3F));
		}
	}

	utf8[length] = 0;
	(*utf8_length) = length;

	return utf8;
}

kuint8* unicode_to_ascii(Unicode* unicode) {
	kuint ascii_length = 0;
	return unicode_to_utf8(unicode, &ascii_length);
}

void build_crc32_table() {
	kuint32 c = 0;
	kuint32 i = i;
	kuint32 j = 0;

	crc32_table = calloc(256, sizeof(kuint32));

	for(i = 0; i < 256; i++) {
		for(c = i << 24, j = 8; j > 0; --j) {
			c = (c & 0x80000000) ? (c << 1) ^ 0x04C11DB7 : (c << 1);
		}
		crc32_table[i] = c;
	}
}

